<?php

namespace SGalinski\SgMail\XClass\Form;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use SGalinski\SgMail\Service\MailTemplateService;
use SGalinski\SgMail\Service\RegisterService;
use SGalinski\SgMail\Service\TypoScriptSettingsService;
use Symfony\Component\Yaml\Yaml;
use TYPO3\CMS\Core\Resource\ResourceStorage;
use TYPO3\CMS\Core\Resource\StorageRepository;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Form\Type\FormDefinitionArray;

/** @noinspection LongInheritanceChainInspection */

/**
 * Displays the form editor. Enables hooking into the save process of the form to handle automatic sg mail registration
 */
class FormEditorController extends \TYPO3\CMS\Form\Controller\FormEditorController {

	/**
	 * This contains type names of fields we dont care about in sgmail
	 *
	 * @var array
	 */
	const IGNORE_FIELDS = [
		'StaticText', 'Hidden', 'GridRow'
	];

	/**
	 * This contains the names of sgmail finishers, so we can identify if we need to generate a template registration
	 *
	 * @var array
	 */
	const MAIL_FINISHER = [
		'MailToSenderFinisher', 'MailToReceiverFinisher'
	];

	/**
	 * Add automatic SgMail Configuration
	 *
	 * @param string $formPersistenceIdentifier
	 * @param FormDefinitionArray $formDefinition
	 * @internal
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 */
	public function saveFormAction(string $formPersistenceIdentifier, FormDefinitionArray $formDefinition) {
		$registrationService = $this->objectManager->get(RegisterService::class);
		/** @noinspection PhpInternalEntityUsedInspection */
		parent::saveFormAction($formPersistenceIdentifier, $formDefinition);

		// immediately exit when no finisher is defined (that means no sgmail integration anyway)
		if (!$formDefinition['finishers']) {
			return;
		}

		/** @var array $finishers */
		$finishers = $formDefinition['finishers'];
		foreach ($finishers as $finisher) {
			// if the current finisher prevents automatic registration, exit this iteration
			if (!$finisher['options']['automaticRegistration']) {
				continue;
			}

			if (!\in_array($finisher['identifier'], self::MAIL_FINISHER, TRUE)) {
				continue;
			}

			// retrieve the extension and template key and jump out of loop
			$extensionKey = (string) $finisher['options']['extension'];
			$extensionKey = $extensionKey ?: 'sg_mail';
			$templateKey = str_replace('_', '-', $finisher['options']['template']);

			// if no template key was explicitly set, use the form identifier as template key
			if ($templateKey === '') {
				$templateKey = $formDefinition['identifier'] . '-' . $finisher['identifier'];
			}

			// if there was no sg mail finisher or an missing key then simply exit the function
			if ($templateKey === '') {
				throw new \InvalidArgumentException(sprintf('Missing template or extension key!'));
			}

			// parse yaml for form fields
			if (strpos($formPersistenceIdentifier, 'EXT:') === 0) {
				$absoluteFilePath = GeneralUtility::getFileAbsFileName($formPersistenceIdentifier);
			} else {
				list($storageUid, $fileIdentifier) = explode(':', $formPersistenceIdentifier, 2);
				$storageRepository = $this->objectManager->get(StorageRepository::class);
				$storage = $storageRepository->findByUid($storageUid);
				if (!$storage instanceof ResourceStorage || !$storage->isBrowsable()) {
					throw new \InvalidArgumentException(
						sprintf('Could not access storage with uid "%d".', $storageUid)
					);
				}
				$absoluteFilePath = $storage->getFile($fileIdentifier);
			}

			$parsedYaml = Yaml::parse(file_get_contents($absoluteFilePath));

			$renderables = [];
			foreach ($parsedYaml['renderables'] as $formPage) {
				if (\is_array($formPage['renderables'])) {
					foreach ($formPage['renderables'] as $row) {
						if (!\is_array($row) || !$row['type']) {
							continue;
						}

						if ($row['type'] === 'GridRow') {
							foreach ($row['renderables'] as $renderableInsideGridrow) {
								$renderables[] = $renderableInsideGridrow;
							}

						} elseif (!\in_array($row['type'], self::IGNORE_FIELDS, TRUE)) {
							$renderables[] = $row;
						}
					}
				} else {
					return;
				}
			}

			// write the new Register.php file and add it to the registrationFiles array in the RegisterService
			$this->writeRegisterFile($renderables, $extensionKey, $templateKey);
		}

		// call register function in register service class
		$registrationService->clearCaches();
		$registrationService->getRegisterArray();
	}

	/**
	 * Builds the register array and saves it to the configured location
	 *
	 * @param array $renderables
	 * @param string $extensionKey
	 * @param string $templateKey
	 * @return string
	 */
	private function writeRegisterFile(array $renderables, $extensionKey, $templateKey): string {
		// get the location where automatic registrations should be stored
		$configurationLocation = $this->getRegistrationPath();

		$registerFolder = GeneralUtility::getFileAbsFileName(
			$configurationLocation
		);

		GeneralUtility::mkdir($registerFolder);

		$hashPrefix = md5($GLOBALS['TYPO3_CONF_VARS']['SYS']['encryptionKey'] . '|' . $templateKey . '.php');
		$registerFile = GeneralUtility::getFileAbsFileName(
			$registerFolder . '/' . $hashPrefix . '_' . $templateKey . '.php'
		);

		// build the register array
		$newRegisterArray = [
			'extension_key' => $extensionKey,
			'template_key' => $templateKey,
			'description' => $templateKey,
			'subject' => $templateKey,
			'markers' => []
		];

		// add the markers for this template
		foreach ($renderables as $element) {
			$markerName = $element['identifier'];
			// if markerName is explicitly set, override the registered identifier
			if (isset($element['properties']['markerName']) && $element['properties']['markerName'] !== '') {
				$markerName = $element['properties']['markerName'];
			}

			$newRegisterArray['markers'][] = [
				'marker' => $markerName,
				'type' => MailTemplateService::MARKER_TYPE_STRING,
				'value' => $element['label'],
				'description' => $element['label']
			];
		}

		file_put_contents($registerFile, '<?php return ' . var_export($newRegisterArray, TRUE) . ';');

		return $registerFile;
	}

	/**
	 * Returns the path to the configured location where automatic mail template registrations should be
	 *
	 * @return string
	 */
	private function getRegistrationPath(): string {
		// get typoscript settings from sg mail
		/** @var TypoScriptSettingsService $typoScriptSettingsService */
		$typoScriptSettingsService = $this->objectManager->get(TypoScriptSettingsService::class);
		$tsSettings = $typoScriptSettingsService->getSettings(0, 'tx_sgmail');

		// get the location where automatic registrations should be stored
		return 'EXT:' . $tsSettings['mail']['configurationLocation'] . '/' . RegisterService::CONFIG_PATH;
	}
}
